import sys


class LineMovement:
    """ A movement to a given line, absolute or relative """

    def __init__(self, absolute=None, relative=None):
        self.absolute = absolute
        self.relative = relative
        if self.absolute and self.relative:
            raise Exception("Cannot move both absolutely and relatively")

    def __add__(self, num):
        if not isinstance(num, type(0)):
            raise Exception("Can only add an integer")
        if self.absolute is not None:
            return self.__class__(absolute=self.absolute + num)
        if self.relative is not None:
            return self.__class__(relative=self.relative + num)

    def do(self, tmux_session):
        if self.relative:
            tmux_session.type_keys("escape", "{}j".format(self.relative))
        elif self.absolute:
            tmux_session.type_keys("escape", "{}G".format(self.absolute))
        self.relative = 0
        self.absolute = None


class VimSession:
    """ A Vim session instrumented through tmux """

    def __init__(self, tmux_session, file_path=None):
        self.tmux_session = tmux_session
        self.file_path = file_path

        self.tmux_session.type_keys("vim")
        if self.file_path:
            self.tmux_session.type_keys(" {}".format(self.file_path))
        self.tmux_session.type_keys("enter")

        self.tmux_session.send_keys(":set paste", "enter")

    def edit_file(self, file_path):
        self.tmux_session.type_keys("escape", ":e ", file_path, "enter")
        self.file_path = file_path

    def apply_patchset(self, patchset):
        for patch in patchset:
            self.apply_patch(patch)

    def apply_patch(self, patch):
        source = patch.source.decode("utf8")
        target = patch.target.decode("utf8")
        if source != target:
            self.tmux_session.type_keys(
                "escape", ":!mv ", source, " ", target, "enter",
            )
        if self.file_path != target:
            self.edit_file(target)

        for hunk in patch:
            self.apply_hunk(hunk)

        self.tmux_session.type_keys("escape", ":w", "enter")

    def write_line(self, line):
        """ Write a line to the vim buffer, assuming everything is set up for it and it
        must be insterted above. """

        if line.startswith(" "):
            lead_spaces = 0
            while lead_spaces < len(line) and line[lead_spaces] == " ":
                lead_spaces += 1
            line = line.strip()

            self.tmux_session.type_keys(
                "O", "escape", "{}a ".format(lead_spaces), "escape", "A", line, "escape"
            )
        else:
            self.tmux_session.type_keys("O", line, "escape")

    def apply_hunk(self, hunk):
        # So far, very naive.
        line_mvt = LineMovement(absolute=hunk.starttgt)

        for b_line in hunk.text:
            line = b_line.decode("utf8")
            if line[0] == " ":
                line_mvt += 1
            else:
                line_mvt.do(self.tmux_session)
                if line[0] == "-":
                    self.tmux_session.type_keys("dd")
                elif line[0] == "+":
                    self.write_line(line.strip()[1:])
                    line_mvt += 1

    def quit(self):
        self.tmux_session.type_keys("escape", ":q", "enter")
